#!/usr/bin/env python2.6

'''This little gem allows to build a photo gallery
including lazy loading of images and a list of thumbnails.'''

import sane
import sys
import os
import hashlib
from bisect import bisect
from threading import Thread
from PyQt4.QtGui import *
from PyQt4.QtCore import *
from types import StringTypes
from lru import LRUCache
from threading import Thread
from Queue import LifoQueue

VERSION = '0.3 (18.04.2009)'

def rint(v):
    return int(round(v))


thumbnailSize = 128

class LoaderThread(Thread):
    def __init__(self, cacheSize=5):
        super(LoaderThread, self).__init__(name='Lazy Pixmap Loader')
        self.daemon = True
        self.requests = LifoQueue()
        
        self.cache = LRUCache(cacheSize)
    
    def add(self, lpm):
        #print 'Adding request to load',lpm.filename
        self.requests.put(lpm)
    
    def register(self, filename, img):
        filename = os.path.abspath(unicode(filename))
        mtime = os.stat(filename).st_mtime
        img.mtime = mtime
        
        self.cache[filename] = img
    
    def run(self):
        while True:
            lpm = self.requests.get()
            filename = os.path.abspath(unicode(lpm.filename))
            mtime = os.stat(filename).st_mtime
            #print 'Lazy Loading',filename
            #print self.cache.dict.keys()
            img = self.cache.get(filename, None)
            if img is None or mtime > img.mtime:
                #print 'Image',filename,'is not in the cache or outdated -> load from disk'
                img = QImage(filename)
                img.mtime = mtime
                # Keep at most letter/A4 sized images at 300DPI in the cache
                if img.width() < 4000 and img.height() < 4000:
                    self.cache[filename] = img
            else:
                #print 'Found image',filename,'in the cache'
                pass
            
            if lpm.scale and not img.isNull():
                #print 'Scaling'
                w,h = lpm.scale
                img = img.scaled(w,h,
                    Qt.AspectRatioMode(Qt.KeepAspectRatio), 
                    Qt.TransformationMode(Qt.SmoothTransformation))
                #print 'Scaled'
            
            lpm.img = img
            lpm.emit(LazyPixmap.loaded, lpm)

loaderThread = LoaderThread()
loaderThread.start()

class LazyPixmap(QObject):
    loaded = SIGNAL('loaded')
    
    def __init__(self, filename, placeholder, scale=None):
        super(LazyPixmap, self).__init__()
        self.filename = filename
        self.scale = scale
        
        if filename is None:
            self.pm = QPixmap()
        else:
            self.pm = placeholder
        
        self.img = None
    
    def load(self):
        if self.filename:
            loaderThread.add(self)
        else:
            self.emit(LazyPixmap.loaded, self)
    
    def getPixmap(self):
        if not self.filename or not self.img:
            return QPixmap()
        
        #print 'Creating pixmap'
        pm = QPixmap.fromImage(self.img)
        #print 'Created pixmap'
        return pm

class PixmapWidget(QWidget):
    zoomFactorChanged = SIGNAL('zoomFactorChanged')
    moveTo = SIGNAL('moveTo(int, int)')
    
    ZOOM_FIT_WINDOW = u'Fit into window'
    ZOOM_FIT_WIDTH = u'Fit width'
    ZOOM_FIT_HEIGHT = u'Fit height'
    
    def __init__(self, filename, name=None, parent=None):
        super(PixmapWidget, self).__init__(parent)
        
        self.name = name
        self._zoomFactor = float(1.)
        self.zooming = False
        self.autoZoomMode = self.ZOOM_FIT_WINDOW
        
        self.setCursor(QCursor(Qt.OpenHandCursor))
        self.cachedPM = None
        self.pm = QPixmap()
        self.text = u'No Image'
        self.left, self.top = 0,0
        
        self.filename = filename
        self.setMinimumSize(128, 128)
    
    def setFilename(self, filename):
        self.left = 0 if self.pm.isNull() else float(-self.x()) / self.pm.width()
        self.top = 0 if self.pm.isNull() else float(-self.y()) / self.pm.height()
        
        self._filename = filename
        self.pm = QPixmap()
        self.lpm = LazyPixmap(filename, self.pm)
        self.text = u'Loading...' if filename else u'No Image'
        self.cachedPM = None
        
        self.connect(self.lpm, LazyPixmap.loaded, self.lazyLoaded)
        self.lpm.load()
    
    def lazyLoaded(self, lpm):
        pm = lpm.getPixmap()
        #print 'loaded',self.lpm.filename,pm
        if hasattr(self, 'lpm'):
            del self.lpm
        self.setPixmap(pm)
        
    def setPixmap(self, pm):
        if pm is None:
            pm = QPixmap()
        
        if not self.pm.isNull():
            self.left = float(-self.x()) / self.pm.width()
            self.top = float(-self.y()) / self.pm.height()
        
        self.pm = pm
        #self.setMinimumSize(self.pm.width(), self.pm.height())
        
        # Force recalc of zoom
        f = self._zoomFactor
        self._zoomFactor = 0
        self.zoomFactor = f if self.autoZoomMode is None else self.autoZoomMode
        
        #print self.left, self.top
        self.emit(self.moveTo, self.left * pm.width(), self.top * pm.height())
        
    def getFilename(self):
        return self._filename
    
    filename = property(getFilename, setFilename)
    
    def setZoomFactor(self, factor, reposPoint=None):
        if not factor:
            #print 'no factor'
            return
        
        p = self.parent()
        if type(factor) in StringTypes:
            if p and self.pm.width():
                wf = float(p.width()) / self.pm.width()
                hf = float(p.height()) / self.pm.height()
                
                if factor == self.ZOOM_FIT_HEIGHT:
                    f = hf
                elif factor == self.ZOOM_FIT_WIDTH:
                    f = wf
                else:
                    f = min(wf, hf)
            else:
                f = 1
        else:
            f = float(factor)
        
        if self._zoomFactor == f or self.pm.isNull():
            return
        
        if reposPoint:
            xoffset, yoffset = self.getOffset()
            if xoffset and yoffset:
                reposPoint = None
            
        if reposPoint:
            x, y = self.mapToModel(reposPoint.x(), reposPoint.y())
            
            deltaX = reposPoint.x() + self.x()
            deltaY = reposPoint.y() + self.y()
        
        #print 'zoom',self.name,repr(factor),f
        self._zoomFactor = f
        self.autoZoomMode = factor if type(factor) in StringTypes else None
        self.emit(self.zoomFactorChanged, self._zoomFactor, self.autoZoomMode)
        
        w = int(round(self.pm.width()*self.zoomFactor))
        h = int(round(self.pm.height()*self.zoomFactor))
        self.setMinimumSize(w, h)
        
        p = self.parent()
        if p:
            self.zooming = True
            self.resize(p.width(), p.height())
            self.zooming = False
            
            if reposPoint:
                newPos = self.mapFromModel(x, y)
                #print 'moveTo',(x,y), (deltaX, deltaY)
                x, y = newPos[0]-deltaX, newPos[1]-deltaY
                self.emit(self.moveTo, x, y)
        
        self.update()
    
    def getZoomFactor(self):
        return self._zoomFactor
    
    zoomFactor = property(getZoomFactor, setZoomFactor)
    
    def mapToModel(self, x, y):
        xoffset, yoffset = self.getOffset()
        #print 'mapToModel', (x,y), (xoffset, yoffset)
        x = x - xoffset
        y = y - yoffset
        x, y = rint(float(x)/self.zoomFactor), rint(float(y)/self.zoomFactor)
        #print x,y
        return x, y
    
    def mapFromModel(self, x, y):
        xoffset, yoffset = self.getOffset()
        x = rint(float(x) * self.zoomFactor) + xoffset
        y = rint(float(y) * self.zoomFactor) + yoffset
        #print 'mapFromModel',x,y
        return x, y
    
    def xmoveEvent(self, event):
        print 'move',event.pos()
        #print
        pass
    
    def getOffset(self):
        xoffset = 0
        yoffset = 0
        
        if self.width() > self.pm.width()*self.zoomFactor:
            xoffset = rint((self.width()-self.pm.width()*self.zoomFactor)/2)
        
        if self.height() > self.pm.height()*self.zoomFactor:
            yoffset = rint((self.height()-self.pm.height()*self.zoomFactor)/2)
        
        return xoffset, yoffset
    
    def paintEvent(self, event):
        xoffset, yoffset = self.getOffset()
        drawBorder = (xoffset or yoffset)
        
        p = QPainter(self)
        
        if self.pm.isNull():
            r = self.rect()
            p.setPen(Qt.black)
            p.drawText(r, Qt.AlignCenter, unicode(self.text))
        else:
            p.save()
            if self.zoomFactor < 1:
                w, h = rint(self.pm.width()*self.zoomFactor), rint(self.pm.height()*self.zoomFactor)
                
                if self.cachedPM and abs(w - self.cachedPM.width()) < 2 and abs(h - self.cachedPM.height()) < 2:
                    pm = self.cachedPM
                else:
                    pm = self.pm.scaled(w, h, 
                        Qt.AspectRatioMode(Qt.KeepAspectRatio), 
                        Qt.TransformationMode(Qt.SmoothTransformation))
                    self.cachedPM = pm
                    #print 'scaling',(w,h),(self.cachedPM.width(),self.cachedPM.height())
                
                p.drawPixmap(xoffset, yoffset, pm)
            else:
                p.translate(xoffset, yoffset)
                p.scale(self.zoomFactor, self.zoomFactor)
                p.drawPixmap(0, 0, self.pm)
            p.restore()
            
            if drawBorder:
                p.setPen(Qt.black)
                p.drawRect(xoffset-1, yoffset-1, 
                    self.pm.width()*self.zoomFactor+1,
                    self.pm.height()*self.zoomFactor+1)
    
    def resizeEvent(self, event):
        if self.zooming:
            return
        
        p = self.parent()
        if not p:
            return
        
        ew = p.width()
        eh = p.height()
        
        pw = rint(self.pm.width()*self.zoomFactor)
        ph = rint(self.pm.height()*self.zoomFactor)
        
        #print 'resizeEvent',(ew,eh),(pw,ph),self.autoZoomMode
        # TODO If a fit mode is selected, always zoom on resize
        if self.autoZoomMode:
            self.zoomFactor = self.autoZoomMode
    
    def wheelEvent(self, event):
        if self.pm.isNull():
            return
        
        f = 0.1
        if event.delta() < 0:
            f = -f
        f = max(self.zoomFactor + f, float(32)/self.pm.width())
        
        self.setZoomFactor(f, event.pos())
    
    def mousePressEvent(self, event):
        pos = event.globalPos()
        self.startPos = pos.x(), pos.y()
        self.startTopLeft = self.x(), self.y()
        #print 'start',self.startPos
        x1, y1 = pos.x(), pos.y()
        x, y = self.mapToModel(x1, y1)
        x2, y2 = self.mapFromModel(x,y)
        #print (x1, y1), (x,y), (x2,y2)
        
        self.setCursor(QCursor(Qt.ClosedHandCursor))
    
    def mouseMoveEvent(self, event):
        pos = event.globalPos()
        x, y = pos.x(), pos.y()
        dx = x - self.startPos[0]
        dy = y - self.startPos[1]
        #print 'move', self.startPos, (x,y), (dx, dy)
        
        x, y = self.startTopLeft[0] + dx, self.startTopLeft[1] + dy
        self.emit(self.moveTo, -x, -y)
        
    def mouseReleaseEvent(self, event):
        pos = event.globalPos()
        x, y = pos.x(), pos.y()
        #print 'stop',(x,y)
        self.setCursor(QCursor(Qt.OpenHandCursor))

class PixmapView(QWidget):
    def __init__(self, name=None, parent=None):
        super(PixmapView, self).__init__(parent)
        
        self.setMinimumSize(256, 256)
        
        self.layout = QVBoxLayout(self)
        self.setLayout(self.layout)
        
        self.scrollArea = QScrollArea(self)
        self.layout.addWidget(self.scrollArea)
        self.scrollArea.setWidgetResizable(True)
        
        self.pw = PixmapWidget(None, name, self.scrollArea)
        self.scrollArea.setWidget(self.pw)
        
        self.statusBar = QWidget(self)
        self.layout.addWidget(self.statusBar)
        
        sbl = QHBoxLayout()
        sbl.setContentsMargins(0, 0, 0, 0)
        self.statusBar.setLayout(sbl)
        
        self.zoomLabel = QLabel(u'&Zoom:')
        sbl.addWidget(self.zoomLabel)
        
        self.zoomCombo = QComboBox(self)
        sbl.addWidget(self.zoomCombo)
        self.zoomLabel.setBuddy(self.zoomCombo)
        
        sbl.addStretch()
        
        self.zoomCombo.setEditable(True)
        self.zoomCombo.setInsertPolicy(QComboBox.NoInsert)
        self.zoomCombo.addItem(PixmapWidget.ZOOM_FIT_WINDOW)
        self.zoomCombo.addItem(PixmapWidget.ZOOM_FIT_WIDTH)
        self.zoomCombo.addItem(PixmapWidget.ZOOM_FIT_HEIGHT)
        self.zoomCombo.addItem(u'25%')
        self.zoomCombo.addItem(u'50%')
        self.zoomCombo.addItem(u'75%')
        self.zoomCombo.addItem(u'100%')
        self.zoomCombo.addItem(u'125%')
        self.zoomCombo.addItem(u'150%')
        self.zoomCombo.addItem(u'200%')
        self.zoomCombo.addItem(u'300%')
        self.zoomCombo.addItem(u'400%')
        self.zoomCombo.addItem(u'800%')
        
        self.connect(self.pw, PixmapWidget.moveTo, self.moveTo)
        self.connect(self.pw, PixmapWidget.zoomFactorChanged, self.zoomFactorChanged)
        
        self.connect(self.zoomCombo, SIGNAL('editTextChanged(const QString &)'), self.zoomComboChanged)
        self.connect(self.zoomCombo, SIGNAL('activated(int)'), self.zoomComboActivated)
    
    def zoomFactorChanged(self, f, mode):
        # Don't mess with widget when user is editing
        if self.zoomCombo.hasFocus():
            return
        
        if mode:
            f = mode
        else:
            f = '%.0f%%' % (f*100)
        
        self.zoomCombo.setEditText(f)
    
    def zoomComboActivated(self, index):
        self.zoomComboChanged(self.zoomCombo.currentText())
    
    def zoomComboChanged(self, text):
        text = unicode(text)
        if text in (PixmapWidget.ZOOM_FIT_WINDOW, PixmapWidget.ZOOM_FIT_WIDTH, PixmapWidget.ZOOM_FIT_HEIGHT):
            f = text
        else:
            if text[-1] == '%':
                text = text[:-1]
            f = float(text) / 100
        
        #print 'zoomComboChanged',f
        self.zoomFactor = f
    
    def moveTo(self, x, y):
        #print 'topLeft', x, y
        self.scrollArea.horizontalScrollBar().setValue(x)
        self.scrollArea.verticalScrollBar().setValue(y)
    
    def setFilename(self, filename):
        self.pw.filename = filename
    
    def getFilename(self):
        return self.pw.filename
    
    filename = property(getFilename, setFilename)
    
    def setZoomFactor(self, zoomFactor):
        self.pw.zoomFactor = zoomFactor
    
    def getZoomFactor(self):
        return self.pw.zoomFactor
    
    zoomFactor = property(getZoomFactor, setZoomFactor)
    
    def setPixmap(self, pm):
        self.pw.setPixmap(pm)
        self.pw.update()
        
    def getPixmap(self):
        return self.pw.pm

class KDEThumbnailCache(QObject):
    imageChanged = SIGNAL('imageChanged') # filename, QImage
    
    def __init__(self, loadingPixmap, cacheSize=128):
        super(KDEThumbnailCache, self).__init__()
        
        self.loadingPixmap = loadingPixmap
        self.cache = LRUCache(cacheSize)
        
        self.basepath = os.path.join(os.environ['HOME'], '.thumbnails', 'large')
        self.size = 256
    
    def fixPath(self, filename):
        return os.path.abspath(unicode(filename))
    
    def getCachePath(self, filename):
        filename = self.fixPath(filename)
        filename = 'file://' + filename
        m = hashlib.md5()
        m.update(filename.encode('utf-8'))
        
        return os.path.join(self.basepath, m.hexdigest())
    
    def getImage(self, filename, size=None):
        filename = self.fixPath(filename)
        if not os.path.exists(filename):
            return QImage()
        
        cachePath = self.getCachePath(filename)
        needsRefresh = False
        if os.path.exists(cachePath) \
                and os.stat(filename).st_mtime > os.stat(cachePath).st_mtime:
            needsRefresh = True
        
        img = self.cache.get(filename, None)
        if img is not None and not needsRefresh:
            if isinstance(img, LazyPixmap):
                img = img.pm.toImage()
            return self.scale(img, size)
        
        if os.path.exists(cachePath) and not needsRefresh:
            #print 'Found thumbnail in cache',filename
            img = QImage(cachePath)
            self.cache[filename] = img
            return self.scale(img, size)
        
        lpm = LazyPixmap(filename, self.loadingPixmap, (self.size, self.size))
        self.connect(lpm, lpm.loaded, self.loaded)
        self.cache[filename] = lpm
        img = self.scale(lpm.pm.toImage(), size)
        lpm.load()
        return img
    
    def scale(self, img, size):
        if not size or img.isNull():
            return img
        
        return img.scaled(size, size,
            Qt.AspectRatioMode(Qt.KeepAspectRatio), 
            Qt.TransformationMode(Qt.SmoothTransformation))
    
    def loaded(self, lpm):
        cachePath = self.getCachePath(lpm.filename)
        img = lpm.img
        #print 'Create thumbnail',cachePath
        img.save(cachePath, 'PNG')
        self.cache[lpm.filename] = img
        
        self.emit(self.imageChanged, lpm.filename, img)
        del lpm

# Based on http://www.saltycrane.com/blog/2008/01/pyqt4-qitemdelegate-example-with/
class ThumbnailDelegate(QItemDelegate):
    def __init__(self, loadingPixmap, font, parent, cacheSize=128):
        QItemDelegate.__init__(self, parent)
        self.fontMetrics = QFontMetrics(font)
        self.setClipping(False)
        
        self.view = parent
        self.cache = KDEThumbnailCache(loadingPixmap, cacheSize)
        self.connect(self.cache, self.cache.imageChanged, self.imageChanged)
    
    def paint(self, painter, option, index):
        painter.save()
        
        painter.setPen(QPen(Qt.NoPen))
        painter.setBrush(option.palette.window())
        painter.drawRect(option.rect)
        
        value = index.data(Qt.DisplayRole)
        if value.isValid():
            filename = unicode(value.toString())
            pm = self.getThumbnail(filename, index)
            
            x = 2 + thumbnailSize//2 - pm.width()//2 + option.rect.left()
            y = 2 + thumbnailSize//2 - pm.height()//2 + option.rect.top()
            
            filename = os.path.basename(filename)
            w = max(pm.width(), self.fontMetrics.width(filename)) + 4
            tr = QRect(x-2, y+pm.height()+2, w, self.fontMetrics.height()+2)
            if option.state & QStyle.State_Selected:
                painter.setBrush(option.palette.highlight())
                r = QRect(x-2, y-2, pm.width()+4, pm.height()+4)
                painter.drawRect(r)
                
                painter.drawRect(tr)
            
            painter.drawPixmap(x, y, pm)
            
            painter.setPen(QPen(Qt.black))
            painter.drawText(tr, Qt.AlignCenter, filename)
        
        painter.restore()
    
    def getThumbnail(self, filename, index):
        return QPixmap.fromImage(self.cache.getImage(filename, thumbnailSize))
        
    def imageChanged(self, filename, img):
        self.view.viewport().update()
        
class GalleryModel(QAbstractListModel):
    def __init__(self, listdata, font, parent=None, *args):
        """ datain: a list where each item is a row
        """
        super(GalleryModel, self).__init__(parent, *args)
        self.listdata = listdata
        self.fontMetrics = QFontMetrics(font)
        self.size = QVariant(QSize(thumbnailSize+4+100, thumbnailSize+6+self.fontMetrics.height()))

    def rowCount(self, parent=QModelIndex()):
        return len(self.listdata)
    
    def setList(self, listdata):
        self.listdata = listdata
        self.reset()
    
    def add(self, filename):
        n = bisect(self.listdata, filename)
        print 'added',n
        self.beginInsertRows(QModelIndex(), n, n)
        self.listdata.insert(n, filename)
        self.endInsertRows()
        return n
    
    def data(self, index, role):
        if not index.isValid():
            return QVariant()
        
        if role == Qt.DisplayRole or role == Qt.ToolTipRole:
            return QVariant(self.listdata[index.row()])
        elif role == Qt.SizeHintRole:
            return self.size
        else:
            return QVariant()

class Gallery(QListView):
    def __init__(self, files, loadingPixmap, parent=None):
        super(Gallery, self).__init__(parent)
        self.setMinimumWidth(256)
        self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        
        self.files = files
        self.loadingPixmap = loadingPixmap
        
        font = self.font()
        
        self.model = GalleryModel(self.files, font, self)
        self.delegate = ThumbnailDelegate(loadingPixmap, font, self)
        self.setModel(self.model)
        self.setItemDelegate(self.delegate)

class GalleryMain(QMainWindow):
    def __init__(self, parent=None):
        super(GalleryMain,self).__init__(parent)
        self.resize(800,600)
        
        self.loadingPixmap = createLoadingPixmap()
        
        self.center = self.splitter = QSplitter(self)
        self.setCentralWidget(self.center)
        
        self.thumbnails = Gallery([], self.loadingPixmap, self)
        self.center.addWidget(self.thumbnails)
        self.center.setStretchFactor(0, 0)
        self.connect(self.thumbnails.selectionModel(), SIGNAL('selectionChanged(QItemSelection, QItemSelection)'), self.thumbnailSelectionChanged)
        
        self.currentImage = PixmapView('Image', self)
        self.center.addWidget(self.currentImage)
        self.center.setStretchFactor(1, 1)
        
    def setFiles(self, files):
        self.thumbnails.model.setList(files)
        
        last = self.thumbnails.model.index(self.thumbnails.model.rowCount()-1,0)
        if last.isValid():
            self.thumbnails.selectionModel().setCurrentIndex(last,
                QItemSelectionModel.ClearAndSelect)
        
    def listdir(self, path):
        l = os.listdir(path)
        i = 0
        while i < len(l):
            f = l[i].lower()
            if (f[0] != '.'
                and (
                    f[-5:] in ('.tiff', '.jpeg')
                    or f[-4:] in ('.gif', '.png', '.tif', '.jpg')
                )):
                i += 1
            else:
                del l[i]
        
        l.sort()
        return [os.path.join(path, f) for f in l]
    
    def thumbnailSelectionChanged(self, selected, deselected):
        index = self.thumbnails.currentIndex()
        #print 'thumbnailSelectionChanged',index
        filename = None
        if index.isValid():
            value = index.data()
            filename = value.toString()
        
        #print filename
        self.currentImage.filename = filename

def createLoadingPixmap():
    loadingPixmap = QPixmap(thumbnailSize, thumbnailSize)
    
    p = QPainter(loadingPixmap)
    p.setPen(QPen(Qt.NoPen))
    p.setBrush(QBrush(Qt.black))
    p.drawRect(loadingPixmap.rect())
    p.setBrush(QBrush(Qt.white))
    p.drawRect(loadingPixmap.rect().adjusted(2,2,-2,-2))
    p.setPen(QPen(Qt.black))
    p.drawText(4, thumbnailSize//2, u'Loading...')
    
    return loadingPixmap

if __name__ == '__main__':
    app = QApplication(sys.argv)
    
    main = GalleryMain()
    path = os.path.join(os.environ['HOME'], 'Scans', 'Rechnungen', '2009')
    l = main.listdir(path)
    main.setFiles(l)
    main.show()
    
    sys.exit(app.exec_())
