#!/usr/bin/env python2.6

'''Little UI to scan from an HP CM1312nfi MFP.

Usage: $0 directory basename

Will put all the images into "directory" and prefix them with "basename".

You can start the scanner by pressing the "Start scan" on the scanner
of by clicking on the "Scan" button in the UI.
'''

import sane
import sys
import os
import _sane
import re
from sane import SaneDev
from threading import Thread, Event
from Queue import Queue, Empty
import Image
import StringIO
from PyQt4.QtGui import *
from PyQt4.QtCore import *
from gallery import GalleryMain, PixmapView, loaderThread
import httplib
import traceback

VERSION="0.4 (18.04.2009)"

#print sane.get_devices()

# TODO
# - Continue scanning until the ADF is empty
# - Allow to scan a page again
# - Delete one or more pages
# - Insert a page
# - Renumber a page range
# - Reorder pages by drag'n'drop
# - Edit project settings in the UI
# - Load/save project settings
# - Select a recent project
# - Support for double-sided batch scanning
# - Ctrl-Q to quit

class SaveThread(Thread):
    saved = SIGNAL('saved') # filename
    
    def __init__(self):
        super(SaveThread,self).__init__()
        
        self.queue = Queue()
        self._run = True
        self.qt = QObject()
    
    def run(self):
        while self._run:
            job = self.queue.get(True)
            if job is None:
                break
                
            img, name, mode, kw = job
            name = os.path.abspath(unicode(name))
            path, basename = os.path.split(name)
            tmp = os.path.join(path, u'.%s_' % basename)
            print 'Saving image %s...' % tmp
            img.save(tmp, mode, **kw)
            os.rename(tmp, name)
            
            self.qt.emit(self.saved, name, img)
    
    def stop(self):
        if self.isAlive():
            self._run = False
            self.queue.put(None)
            self.join(60)
    
    def save(self, im, name, mode, **kw):
        self.queue.put((im, name, mode, kw))

class ScannerThread(Thread):
    scanStarted = SIGNAL('scanStarted') # self
    newImageAvailable = SIGNAL('newImageAvailable') # self
    
    def __init__(self, dev):
        super(ScannerThread,self).__init__()
        
        self.dev = dev
        self.autoSource = False
        self.waiting = False
    
        if 0:
            print dev.get_parameters()
        
        if 0:
            for opt in dev.get_options():
                print opt
        if 0:
            print dev.optlist
        
        self.qt = QObject()
        
        self.im = None
        self.queue = Queue()
        self.conn = httplib.HTTPConnection(dev.ip)
        
        self.startScanPattern = re.compile(r'<StartScan>(\d)+')
        self.adfLoadedPattern = re.compile(r'<ADFLoaded>(\d)+')
        
    def emit(self, signal, *args):
        self.qt.emit(signal, *args)
    
    def refreshScanOptions(self):
        url = '/hp/device/notifications.xml'
        self.conn.request('GET', url)
        resp = self.conn.getresponse()
        if resp.status != httplib.OK:
            raise Exception('Scanner returned %d %s' % (resp.status, resp.reason))
        
        data = resp.read()
        #print data
        
        m = self.adfLoadedPattern.search(data)
        #print 'adf',m.group(1)
        self.adfLoaded = int(m.group(1))
        if self.autoSource:
            self.dev.source = 'ADF' if self.adfLoaded else 'Flatbed'
        m = self.startScanPattern.search(data)
        #print 'scan',m.group(1)
        self.scanButtonPressed = int(m.group(1))
    
    def scan(self):
        self.queue.put('Scan')
    
    def createImage(self):
        "Snap a picture, returning a PIL image object with the results"
        (mode, last_frame,
         (xsize, ysize), depth, bytes_per_line) = self.dev.get_parameters()
        if mode in ['gray', 'red', 'green', 'blue']:
            format = 'L'
        elif mode == 'color':
            format = 'RGB'
        else:
            raise ValueError('got unknown "mode" from self.get_parameters()')
        self.im = Image.new(format, (xsize,ysize))
        return self.im
    
    def run(self):
        while True:
            try:
                self.refreshScanOptions()
                if self.scanButtonPressed:
                    self.scan()
            except:
                traceback.print_exc()
            
            # Wait for start signal
            try:
                s = self.queue.get(True, 1)
            except Empty:
                continue
            
            if not s:
                break
            
            try:
                self.scanOneFrame()
            except:
                traceback.print_exc()
    
    def stop(self):
        if self.isAlive():
            self.queue.put(None)
            self.join(60)
    
    def scanOneFrame(self):
        print 'Getting scan'
        needsCancel = False
        self.waiting = False
        try:
            self.createImage()
            print 'emit scanStarted'
            self.emit(self.scanStarted, self)
            
            try:
                needsCancel = True
                self.dev.start()
                self.dev.dev.snap(self.im.im.id, 1)
            except _sane.error, v:
                print v
                print repr(v)
                if v == 'Document feeder out of documents':
                    pass
                elif v == 'Operation was cancelled':
                    pass
                else:
                    raise Exception(v)
            finally:
                if dev.source != 'ADF':
                    needsCancel = False
            
            print 'emit newImageAvailable'
            self.emit(self.newImageAvailable, self)
            #print 'Done.'
        finally:
            self.refreshScanOptions()
            if self.adfLoaded:
                self.waiting = True
                return
            
            if needsCancel:
                self.dev.cancel()

def pilImage2QImage(im, width=None, height=None):
    if im is None:
        return QPixmap()
    
    if 0: # Uses less memory but slow or ugly
        if width or height:
            if width and not height:
                height = width
            elif height and not width:
                width = height
            
            im = im.copy()
            # Slow :(
            #im.thumbnail((width, height), Image.ANTIALIAS)
            # Ugly :(
            #im.thumbnail((width, height), Image.BILINEAR)
            im.thumbnail((width, height), Image.BICUBIC)
    
    b = im.convert('RGB').tostring()
    #print len(b), repr(b[:32])
    w, h = im.size
    #print w*h*3
    #print w*h*4
    bytesPerLine = len(b)//h
    img = QImage(b, w, h, bytesPerLine, QImage.Format_RGB888)
    img.b = b
    
    if 1: # Needs more RAM but fast
        if width and height:
            img = img.scaled(width, height,
                Qt.KeepAspectRatio,
                Qt.SmoothTransformation)
        elif height:
            img = img.scaledToHeight(height, Qt.SmoothTransformation)
        elif width:
            img = img.scaledToWidth(width, Qt.SmoothTransformation)
    
    return img

class Main(GalleryMain):
    def __init__(self, dev, project, parent=None):
        super(Main,self).__init__(parent)
        self.resize(1200,600)
        
        self.dev = dev
        self.project = project
        self.project.updateSaneDevice(self.dev)
        
        self.toolbar = QToolBar()
        self.addToolBar(self.toolbar)
        
        self.currentScan = PixmapView('Scan', self)
        self.splitter.addWidget(self.currentScan)
        self.splitter.setStretchFactor(2, 1)
        
        self.timer = QTimer()
        self.timer.setInterval(1000)
        self.connect(self.timer, SIGNAL('timeout()'), self.tick)
        
        self.setFiles(self.project.listdir())
        
        self.scanner = False
        
        self.saveThread = SaveThread()
        self.connect(self.saveThread.qt, SaveThread.saved, self.saved)
        
        self.scannerThread = ScannerThread(self.dev)
        self.scannerThread.autoSource = self.project.autoSource
        self.connect(self.scannerThread.qt, ScannerThread.newImageAvailable, self.newImageAvailable)
        self.connect(self.scannerThread.qt, ScannerThread.scanStarted, self.scanStarted)
        
        self.createActions()
        self.toolbar.addAction(self.startScanAction)
    
    def startThreads(self):
        self.saveThread.start()
        self.scannerThread.start()
    
    def stopThreads(self):
        self.saveThread.stop()
        self.scannerThread.stop()
    
    def createActions(self):
        self.startScanAction = QAction(u'&Scan', self)
        self.connect(self.startScanAction, SIGNAL('triggered(bool)'), self.startScan)
    
    def startScan(self, checked=False):
        self.scannerThread.scan()
    
    def scanStarted(self, st):
        print 'scanStarted'
        self.filename = self.project.next()
        self.scanner = True
        self.startScanAction.setEnabled(False)
    
    def newImageAvailable(self, st):
        print 'newImageAvailable'
        self.scanner = False
        self.refreshCurrentScan()
        im = st.im
        self.startScanAction.setEnabled(True)
        
        self.saveThread.save(im, self.filename, self.project.format, **self.project.saveOptions)
    
    def saved(self, filename, im):
        print 'saved',repr(filename)
        img = pilImage2QImage(im)
        loaderThread.register(filename, img)
        
        # More documents in the ADF feeder
        if self.scannerThread.waiting:
            self.scannerThread.scan()
        
        model = self.thumbnails.model
        n = model.add(filename)
        index = model.index(n, 0)
        if index.isValid():
            self.thumbnails.selectionModel().setCurrentIndex(index,
                QItemSelectionModel.ClearAndSelect)
    
    def refreshCurrentScan(self):
        #print 'refreshCurrentScan'
        im = self.scannerThread.im
        size = self.currentScan.scrollArea.viewport().size()
        img = pilImage2QImage(im, size.width(), size.height())
        pm = QPixmap.fromImage(img)
        self.currentScan.setPixmap(pm)
    
    def show(self):
        self.timer.start()
        QMainWindow.show(self)
    
    def tick(self, *args):
        s = 'Inactive'
        if self.scanner:
            self.refreshCurrentScan()
            s = u'Scanning %s' % self.filename
        elif self.currentImage.filename is not None:
            s =  unicode(self.currentImage.filename)
        
        self.setWindowTitle('PyScan %s - %s' % (VERSION, s))

class ScanProject(object):
    def __init__(self, path, prefix=u'Seite', format='TIFF'):
        self.path = os.path.abspath(path)
        self.mode = 'Color'
        self.resolution = 300
        self.source = 'Auto'
        self.format = format
        self.saveOptions = dict(
                quality=90,
        )
        self.start = 1
        self.step = 1
        self.prefix = unicode(prefix)
        if self.prefix[-1] not in '.-_':
            self.prefix += '-'
        
        if self.format == 'JPEG':
            self.suffix = u'.jpg'
        else:
            self.suffix = u'.%s' % (self.format.lower())
        
        self.autoSource = self.source == 'Auto'
        self.refresh()
    
    def isAutoSource(self):
        return self.source == 'Auto'
    
    def updateSaneDevice(self, dev):
        dev.mode = self.mode
        dev.resolution = self.resolution
        if not self.autoSource:
            dev.source = self.source
    
    def add(self, filename):
        self.files.append(filename)
        self.files.sort()
    
    def refresh(self):
        self.files = None
        self.listdir()
    
    def listdir(self):
        if self.files:
            return self.files
        
        pattern = re.compile(re.escape(self.prefix)+'(\d{4})\.[a-zA-Z]{3,4}')
        
        l = []
        n = 1
        for f in os.listdir(self.path):
            m = pattern.match(f)
            if m:
                l.append(f)
                n = max(n, int(m.group(1)))
        l.sort()
        self.files = [os.path.join(self.path, f) for f in l]
        if n:
            self.start = n
        return self.files
    
    def getFilename(self, n):
        return os.path.join(self.path, u'%s%04d%s' % (self.prefix, n, self.suffix))
    
    def last(self):
        l = self.listdir()
        return l[-1] if l else None
    
    def next(self):
        n = self.start
        
        while True:
            filename = self.getFilename(n)
            if not os.path.exists(filename):
                break
            n += self.step
        
        self.start = n
        return filename


if __name__ == '__main__':
    if 1:
        app = QApplication(sys.argv)
        
        ip = '192.168.1.242'
        uri = 'hpaio:/net/HP_Color_LaserJet_CM1312nfi_MFP?ip=%s' % ip
        try:
            dev = SaneDev(uri, False)
        except TypeError, v:
            if str(v).startswith('__init__() takes at most'):
                raise Exception("It seems you didn't apply the patch in sane-py.patch (see README)")
            else:
                raise
        except _sane.error, v:
            print v, repr(v)
        dev.ip = ip
        dev.compression = 'None'

        #filePattern = sys.argv[1]
        #start = int(sys.argv[2])
        #inc = int(sys.argv[3])
        
        path = sys.argv[1]
        if not os.path.exists(path):
            os.makedirs(path)
        elif not os.path.isdir(path):
            raise Exception('%s ist kein Verzeichnis' % path)
        
        prefix = 'Seite'
        if len(sys.argv) >= 3:
            prefix = sys.argv[2]
        format = 'TIFF'
        if len(sys.argv) >= 4:
            format = sys.argv[3].upper()
            if format == 'JPG':
                format = 'JPEG'
        
        project = ScanProject(path, prefix=prefix, format=format)
        
        main = Main(dev, project)
        main.startThreads()
        #main.setScanner(dev, filePattern, start, inc)
        main.show()
        
        rc = app.exec_()
        main.stopThreads()
        dev.close()
        
        sys.exit(rc)
    
    if 0:
        im = Image.open('../scanner.bak/test.tiff')
        img = pilImage2QImage(im)
        img.save('qt-test.tiff', 'TIFF')
